#include "script.h"
#include "grub_script.tab.h"

struct grub_script_node *
grub_script_newnode(struct grub_parser_param *state,grub_script_node_type_t type)
{
  struct grub_script_node *res;

  res=(struct grub_script_node *)grub_malloc(sizeof(*res));
  res->type=type;
  res->next=res->child=0;
  res->flst=state->free_list;
  state->free_list=res;
  state->mem_alloc+=sizeof(*res);
  return res;
}

struct grub_script_node *
grub_script_newstr(struct grub_parser_param *state,grub_script_node_type_t type,char* str)
{
  struct grub_script_node *res;
  int len;

  res=(struct grub_script_node *)grub_malloc(sizeof(*res));
  res->type=type;
  len=grub_strlen(str);
  res->child=(struct grub_script_node *)grub_malloc(len+1);
  res->next=0;
  res->flst=state->free_list;
  state->free_list=res;
  grub_strcpy((char*)res->child,str);
  state->mem_alloc+=sizeof(*res)+len+1;
  return res;
}

struct grub_script_node *
grub_script_getnode(struct grub_parser_param *state,struct grub_script_node *node)
{
  if (! node)
    return 0;

  if (node==state->free_list)
    {
      state->free_list=node->flst;
    }
  else
    {
      struct grub_script_node *head;

      head=state->free_list;
      while (head)
        {
          if (head->flst==node)
            break;
          head=head->flst;
        }
      if (head)
          head->flst=node->flst;
      else
        grub_printf("Warning: node not found in free list\n");
    }
  node->flst=0;
  return node;
}

void
grub_script_freenode(struct grub_parser_param *state,struct grub_script_node *node)
{
  if (node->type<=GRUB_SCRIPT_NODE_TYPE_VAR)
    {
      state->mem_alloc-=grub_strlen((char*)node->child)+1;
      grub_free(node->child);
      if (node->next)
        grub_script_freenode(state,node->next);
    }
  else
    {
      if (node->child)
        grub_script_freenode(state, node->child);
      if (node->next)
        grub_script_freenode(state, node->next);
    }
  grub_free(node);
  state->mem_alloc-=sizeof(*node);
}

struct grub_script_node *
grub_script_catnode(struct grub_parser_param *state,struct grub_script_node *head,struct grub_script_node *tail)
{
  struct grub_script_node *p;

  if (! head)
    return tail;

  if (! tail)
    return head;

  for (p=head;p->next;)
    p=p->next;

  p->next=grub_script_getnode(state,tail);

  return head;
}

int flag;

grub_err_t getline (char **line)
{
  char buf[256];

  if (flag)
    {
      if (flag==1)
        {
          grub_printf("## ");
          flag++;
        }
      else
        grub_printf(">> ");
    }
back:
  if (fgets(buf,sizeof(buf),stdin))
    {
      char *pb;
      int len;

      pb=grub_malloc(grub_strlen(buf)+1);
      grub_strcpy(pb,buf);
      len=grub_strlen(pb)-1;
      while ((len>=0) &&
             ((pb[len]=='\n') || (pb[len]=='\r')))
        len--;
      if (len<0)
        goto back;
      pb[len+1]=0;
      *line=pb;
    }
  else
    *line=0;
  return 0;
}

void grub_script_free(struct grub_parser_param *state)
{
  struct grub_script_node *p;

  if (state->root)
    grub_script_freenode(state, state->root);
  p=state->free_list;
  while (p)
    {
      struct grub_script_node *s;

      s=p;
      p=p->flst;
      grub_script_freenode (state, s);
    }
  if (state->mem_alloc)
    grub_printf("Warning: memory leak (%d bytes)\n",state->mem_alloc);
  state->mem_alloc=0;
  state->free_list=0;
  grub_free(state->lexerstate->buffer);
  grub_free(state->lexerstate);
}

void dump_tree(struct grub_script_node *node,int level)
{
  int i;

  for (i=0;i<level;i++)
    printf("  ");
  switch (node->type)
    {
    case GRUB_SCRIPT_NODE_TYPE_STR:
      printf("STR");
      break;
    case GRUB_SCRIPT_NODE_TYPE_VAR:
      printf("VAR");
      break;
    case GRUB_SCRIPT_NODE_TYPE_TEXT:
      printf("TEXT");
      break;
    case GRUB_SCRIPT_NODE_TYPE_IF:
      printf("IF");
      break;
    case GRUB_SCRIPT_NODE_TYPE_CMD:
      printf("CMD");
      break;
    case GRUB_SCRIPT_NODE_TYPE_CMDS:
      printf("CMDS");
      break;
    case GRUB_SCRIPT_NODE_TYPE_FUNC:
      printf("FUNC");
      break;
    case GRUB_SCRIPT_NODE_TYPE_MENU:
      printf("MENU");
    }
  if (node->type<=GRUB_SCRIPT_NODE_TYPE_VAR)
    printf("  \"%s\"\n",node->child);
  else
    {
      printf("\n");
      if (node->child)
        dump_tree(node->child,level+1);
    }
  if (node->next)
    dump_tree(node->next, level);

}

int
main(int argc,char** argv)
{
  char* line;
  struct grub_parser_param state;

  memset(&state, 0, sizeof(state));

  flag=(! ((argc>=2) && (argv[1][0]=='-')));
  while (1)
    {
      struct grub_script_node *script;
      if (flag)
        flag=1;
      getline(&line);
      if (! line)
        break;
      state.lexerstate=grub_script_lexer_init(line, getline);
      state.root=0;
      grub_script_yyparse(&state);
      if (state.root)
        {
          grub_printf("== TREE BEGIN ==\n");
          dump_tree(state.root,0);
          grub_printf("== TREE END ==\n");
        }
      grub_script_free(&state);
      grub_free(line);
    }
  return 0;
}
