/* parser.y - The scripting parser.  */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2005, 2006  Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

%{
#include "script.h"
//#include <grub/mm.h>

#define YYFREE		grub_free
#define YYMALLOC	grub_malloc
#define YYLTYPE_IS_TRIVIAL      0

%}

%union {
  struct grub_script_node *node;
}

%token GRUB_PARSER_TOKEN_IF
%token GRUB_PARSER_TOKEN_WHILE
%token GRUB_PARSER_TOKEN_FUNCTION
%token GRUB_PARSER_TOKEN_MENUENTRY
%token GRUB_PARSER_TOKEN_ELSE
%token GRUB_PARSER_TOKEN_THEN
%token GRUB_PARSER_TOKEN_FI
%token GRUB_PARSER_TOKEN_TEXT
%token GRUB_PARSER_TOKEN_ERROR

%type <node> GRUB_PARSER_TOKEN_TEXT text command commands simple_cmd if_command commandblock function menuentry

%pure-parser
%lex-param { struct grub_parser_param *state };
%parse-param { struct grub_parser_param *state };

%%

script:		commands
                {
                  state->root=grub_script_getnode(state, $1);
                }
                | menuentry delimiter
                {
                  state->root=grub_script_getnode(state, $1);
                }
                | function delimiter
                {
                  state->root=grub_script_getnode(state, $1);
                }

delimiter:	'\n'
		| ';'
		| delimiter '\n'

newlines:
		| newlines '\n'

text		: GRUB_PARSER_TOKEN_TEXT
		{
		  $$=$1;
		}
		| GRUB_PARSER_TOKEN_IF
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "if"));
		}
		| GRUB_PARSER_TOKEN_WHILE
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "while"));
		}
		| GRUB_PARSER_TOKEN_FUNCTION
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "function"));
		}
		| GRUB_PARSER_TOKEN_MENUENTRY
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "menuentry"));
		}
		| GRUB_PARSER_TOKEN_ELSE
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "else"));
		}
		| GRUB_PARSER_TOKEN_THEN
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "then"));
		}
		| GRUB_PARSER_TOKEN_FI
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "fi"));
		}
		| '{'
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "{"));
		}
		| '}'
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_TEXT);
		  $$->child=grub_script_getnode(state, grub_script_newstr(state, GRUB_SCRIPT_NODE_TYPE_STR, "}"));
		}
		| GRUB_PARSER_TOKEN_ERROR
		{
		  $$=0;
		  yyerror(state, "invalid token");
		}

command:	simple_cmd delimiter
		{
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_CMD);
		  $$->child=grub_script_getnode(state, $1);
		}
		| commandblock delimiter
		{
		  $$=$1;
		}
		| if_command delimiter
		{
		  $$=$1;
		}
		| '}' text delimiter
		{
		  $$=0;
		  yyerror(state, "} without {");
		}
		| GRUB_PARSER_TOKEN_WHILE text delimiter
		{
		  $$=0;
		  yyerror(state, "while not implemented");
		}
		| ';' newlines
		{
                  $$=0;
                  yyerror(state, "empty command");
		}
		| GRUB_PARSER_TOKEN_ERROR text delimiter
		{
		  $$=0;
		  yyerror(state, "invalid leading token");
		}
		| error delimiter
		{
		  $$=0;
		  yyerror(state, "invalid command");
		  yyerrok;
		}

commands:	command
		{
		  $$=$1;
		}
		| commands command
		{
		  $$=grub_script_catnode(state,$1,$2);
		}

simple_cmd:	GRUB_PARSER_TOKEN_TEXT
		{
		  $$=$1;
		}
		| simple_cmd text
		{
		  $$=grub_script_catnode(state,$1,$2);
		}

if_start:	GRUB_PARSER_TOKEN_IF
		{
		  grub_script_lexer_ref (state->lexerstate);
		}

if_command:	if_start newlines commands GRUB_PARSER_TOKEN_THEN newlines commands GRUB_PARSER_TOKEN_FI
		{
		  struct grub_script_node *p;

		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_IF);
		  $$->child=grub_script_getnode(state, grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_CMDS));
		  p=$$->child;
		  p->child=grub_script_getnode(state,$3);
		  p->next=grub_script_getnode(state, grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_CMDS));
		  p=p->next;
		  p->child=grub_script_getnode(state,$6);
		  grub_script_lexer_deref (state->lexerstate);
		}
		| if_start newlines commands GRUB_PARSER_TOKEN_THEN newlines commands
		GRUB_PARSER_TOKEN_ELSE newlines commands GRUB_PARSER_TOKEN_FI
		{
		  struct grub_script_node *p;

		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_IF);
		  p=$$->child=grub_script_getnode(state, grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_CMDS));
		  p->child=grub_script_getnode(state,$3);
		  p->next=grub_script_getnode(state, grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_CMDS));
		  p=p->next;
		  p->child=grub_script_getnode(state,$6);
		  p->next=grub_script_getnode(state, grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_CMDS));
		  p=p->next;
		  p->child=grub_script_getnode(state,$9);
		  grub_script_lexer_deref (state->lexerstate);
		}
		| if_start newlines error GRUB_PARSER_TOKEN_FI
		{
		  grub_script_lexer_deref (state->lexerstate);
		  $$=0;
		  yyerror(state, "invalid if block");
		  yyerrok;
		}

cb_start:	'{'
		{
		  grub_script_lexer_ref (state->lexerstate);
		}

commandblock:	cb_start newlines commands '}'
		{
                  grub_script_lexer_deref (state->lexerstate);
                  $$ = grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_CMDS);
                  $$->child = grub_script_getnode(state, $3);
		}
		| cb_start newlines commands error '}'
		{
		  grub_script_lexer_deref (state->lexerstate);
                  $$ = grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_CMDS);
                  $$->child = grub_script_getnode(state, $3);
		  yyerror(state, "missing ending delimiter in command block");
		  yyerrok;
		}
		| cb_start newlines error '}'
		{
		  grub_script_lexer_deref (state->lexerstate);
                  $$ = 0;
		  yyerror(state, "invalid command block");
		  yyerrok;
		}

mn_start:	GRUB_PARSER_TOKEN_MENUENTRY
		{
		  grub_script_lexer_ref (state->lexerstate);
		}

menuentry:	mn_start GRUB_PARSER_TOKEN_TEXT newlines commandblock
		{
		  grub_script_lexer_deref (state->lexerstate);
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_MENU);
		  $$->child=grub_script_getnode(state, $2);
		  $$->child->next=grub_script_getnode(state, $4);
		}
		| mn_start error
		{
		  grub_script_lexer_deref (state->lexerstate);
		  $$=0;
		  yyerror(state, "invalid menuentry");
		  yyerrok;
		}

fn_start:	GRUB_PARSER_TOKEN_FUNCTION
		{
		  grub_script_lexer_ref (state->lexerstate);
		}

function:	fn_start GRUB_PARSER_TOKEN_TEXT newlines commandblock
		{
		  grub_script_lexer_deref (state->lexerstate);
		  $$=grub_script_newnode(state, GRUB_SCRIPT_NODE_TYPE_FUNC);
		  $$->child=grub_script_getnode(state, $2);
		  $$->child->next=grub_script_getnode(state, $4);
		}
		| fn_start error
		{
		  grub_script_lexer_deref (state->lexerstate);
		  $$=0;
		  yyerror(state, "invalid function");
		  yyerrok;
		}
