#ifndef __SCRIPT_H
#define __SCRIPT_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define grub_free	free
#define grub_malloc	malloc
#define grub_strcpy	strcpy
#define grub_strlen	strlen
#define grub_strcmp	strcmp
#define grub_printf	printf

typedef int grub_err_t;

typedef enum
{
  GRUB_SCRIPT_NODE_TYPE_STR,
  GRUB_SCRIPT_NODE_TYPE_VAR,
  GRUB_SCRIPT_NODE_TYPE_TEXT,
  GRUB_SCRIPT_NODE_TYPE_IF,
  GRUB_SCRIPT_NODE_TYPE_CMD,
  GRUB_SCRIPT_NODE_TYPE_CMDS,
  GRUB_SCRIPT_NODE_TYPE_FUNC,
  GRUB_SCRIPT_NODE_TYPE_MENU,
  GRUB_SCRIPT_NODE_TYPE_ERROR,
} grub_script_node_type_t;

struct grub_script_node;

struct grub_script_node {
  grub_script_node_type_t type;
  struct grub_script_node *child;
  struct grub_script_node *next;
  struct grub_script_node *flst;
};

typedef enum
{
  GRUB_PARSER_STATE_QUOTE=1,
  GRUB_PARSER_STATE_DQUOTE,
  GRUB_PARSER_STATE_TEXT,
  GRUB_PARSER_STATE_VAR,
  GRUB_PARSER_STATE_VARNAME,
  GRUB_PARSER_STATE_VARNAME2,
  GRUB_PARSER_STATE_QVAR,
  GRUB_PARSER_STATE_QVARNAME,
  GRUB_PARSER_STATE_QVARNAME2
} grub_parser_state_t;

/* State of the lexer as passed to the lexer.  */
struct grub_lexer_param
{
  /* Set to 0 when the lexer is done.  */
  int done;

  /* State of the state machine.  */
  grub_parser_state_t state;

  /* Function used by the lexer to get a new line when more input is
     expected, but not available.  */
  grub_err_t (*getline) (char **);

  /* A reference counter.  If this is >0 it means that the parser
     expects more tokens and `getline' should be called to fetch more.
     Otherwise the lexer can stop processing if the current buffer is
     depleted.  */
  int refs;

  /* The character stream that has to be parsed.  */
  char *script;
  char *newscript; /* XXX */

  char read_ahead;	/* read ahead char */
  char *buffer;		/* token buffer */
  int offset;		/* token buffer pointer */
};

struct grub_parser_param{
  int mem_alloc;
  struct grub_script_node *free_list;
  struct grub_lexer_param *lexerstate;
  struct grub_script_node *root;
};

struct grub_script_node *grub_script_newnode(struct grub_parser_param *,grub_script_node_type_t type);
struct grub_script_node *grub_script_newstr(struct grub_parser_param *,grub_script_node_type_t type,char* str);
struct grub_script_node *grub_script_getnode(struct grub_parser_param *,struct grub_script_node *node);
void grub_script_freenode(struct grub_parser_param *,struct grub_script_node *node);
struct grub_script_node *grub_script_catnode(struct grub_parser_param *,struct grub_script_node *head,struct grub_script_node *tail);

struct grub_lexer_param *grub_script_lexer_init (char *script, grub_err_t (*getline) (char **));

void dump_tree(struct grub_script_node *node,int level);

#endif /* __SCRIPT_H */
